from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID

from .. import types
from ..program_id import PROGRAM_ID


class CreateArgs(typing.TypedDict):
    create_args: types.create_args.CreateArgsKind


layout = borsh.CStruct("create_args" / types.create_args.layout)


class CreateAccounts(typing.TypedDict):
    metadata: Pubkey
    master_edition: typing.Optional[Pubkey]
    mint: Pubkey
    authority: Pubkey
    payer: Pubkey
    update_authority: Pubkey
    sysvar_instructions: Pubkey
    spl_token_program: typing.Optional[Pubkey]


def create(
    args: CreateArgs,
    accounts: CreateAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["metadata"], is_signer=False, is_writable=True),
        (
            AccountMeta(
                pubkey=accounts["master_edition"], is_signer=False, is_writable=True
            )
            if accounts["master_edition"]
            else AccountMeta(pubkey=program_id, is_signer=False, is_writable=False)
        ),
        AccountMeta(pubkey=accounts["mint"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["authority"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["payer"], is_signer=True, is_writable=True),
        AccountMeta(
            pubkey=accounts["update_authority"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["sysvar_instructions"], is_signer=False, is_writable=False
        ),
        (
            AccountMeta(
                pubkey=accounts["spl_token_program"], is_signer=False, is_writable=False
            )
            if accounts["spl_token_program"]
            else AccountMeta(pubkey=program_id, is_signer=False, is_writable=False)
        ),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts

    # DON 29-Jan-2024 - TODO_ORIGINAL The original code uses this byte string as the identifier, however it doesn't work
    # We use the below descriminator of 42 instead which is taken from the javascript SDK.
    # identifier = b"\x18\x1e\xc8(\x05\x1c\x07w"
    identifier = bytes([42])
    encoded_args = layout.build(
        {
            "create_args": args["create_args"].to_encodable(),
        }
    )
    data = identifier + encoded_args

    return Instruction(program_id, data, keys)
